#ifndef _GBA_BACKGROUND_H
#define _GBA_BACKGROUND_H

//////////////////////////////////////////////////////////////////////
// TYPES ET DEFINITIONS
////////////////////////////////////////////////////////////////////// 

#define CHAR_SHIFT            2
#define SCREEN_SHIFT          8
#define TEXTBG_SIZE_256x256   0x0
#define TEXTBG_SIZE_256x512   0x8000
#define TEXTBG_SIZE_512x256   0x4000
#define TEXTBG_SIZE_512x512   0xC000

#define ROTBG_SIZE_128x128    0x0
#define ROTBG_SIZE_256x256    0x4000
#define ROTBG_SIZE_512x512    0x8000
#define ROTBG_SIZE_1024x1024  0xC000

#define BG_WRAPAROUND         0x1

#define BG_MOSAIC_ENABLE      0x40
#define BG_COLOR_256          0x80
#define BG_COLOR_16           0x0

#define REG_BG0CNT     *(u16*) 0x4000008    // Background 0
#define REG_BG1CNT     *(u16*) 0x400000A    // Background 1
#define REG_BG2CNT     *(u16*) 0x400000C    // Background 2
#define REG_BG3CNT     *(u16*) 0x400000E    // Background 3

#define REG_BG0HOFS    *(u16*) 0x4000010    // Background 0 Horizontal Offset
#define REG_BG0VOFS    *(u16*) 0x4000012    // Background 0 Vertical Offset
#define REG_BG1HOFS    *(u16*) 0x4000014    // Background 1 Horizontal Offset
#define REG_BG1VOFS    *(u16*) 0x4000016    // Background 1 Vertical Offset
#define REG_BG2HOFS    *(u16*) 0x4000018    // Background 2 Horizontal Offset
#define REG_BG2VOFS    *(u16*) 0x400001A    // Background 2 Vertical Offset
#define REG_BG3HOFS    *(u16*) 0x400001C    // Background 3 Horizontal Offset
#define REG_BG3VOFS    *(u16*) 0x400001E    // Background 3 Vertical Offset
#define REG_BG2PA      *(u16*) 0x4000020    // Background 2 PA Rotation (pa = x_scale * cos(angle);)
#define REG_BG2PB      *(u16*) 0x4000022    // Background 2 PB Rotation (pb = y_scale * sin(angle);)
#define REG_BG2PC      *(u16*) 0x4000024    // Background 2 PC Rotation (pc = x_scale * -sin(angle);)
#define REG_BG2PD      *(u16*) 0x4000026    // Background 2 PD Rotation (pd = y_scale * cos(angle);)
#define REG_BG2X       *(u32*) 0x4000028    // Background 2 X Location
#define REG_BG2X_L     *(u16*) 0x4000028    // Background 2 Starting point low
#define REG_BG2X_H     *(u16*) 0x400002A    // Background 2 Starting point high
#define REG_BG2Y       *(u32*) 0x400002C    // Background 2 Y Location
#define REG_BG2Y_L     *(u16*) 0x400002C    // Background 2 Starting point low
#define REG_BG2Y_H     *(u16*) 0x400002E    // Background 2 Starting point high
#define REG_BG3PA      *(u16*) 0x4000030    // Background 3 PA Rotation (pa = x_scale * cos(angle);)
#define REG_BG3PB      *(u16*) 0x4000032    // Background 3 PB Rotation (pb = y_scale * sin(angle);)
#define REG_BG3PC      *(u16*) 0x4000034    // Background 3 PC Rotation (pc = x_scale * -sin(angle);)
#define REG_BG3PD      *(u16*) 0x4000036    // Background 3 PD Rotation (pd = y_scale * cos(angle);)
#define REG_BG3X       *(u32*) 0x4000038    // Background 3 X Location
#define REG_BG3X_L     *(u16*) 0x4000038    // Background 3 Starting point low
#define REG_BG3X_H     *(u16*) 0x400003A    // Background 3 Starting point high
#define REG_BG3Y       *(u32*) 0x400003C    // Background 3 Y Location
#define REG_BG3Y_L     *(u16*) 0x400003C    // Background 3 Starting point low
#define REG_BG3Y_H     *(u16*) 0x400003E    // Background 3 Starting point high

#define REG_MOSAIC     *(u32*) 0x400004C    // Mosaic Mode

// MACROS
// 0x6000000 dsigne l'adresse du videoFrontBuffer
#define CharBaseBlock(n)      (((n) * 0x4000) + 0x6000000)
#define ScreenBaseBlock(n)    (((n) * 0x800)  + 0x6000000)

#define MOS_BG_HOR(n)             (n)                  // mosaque
#define MOS_BG_VER(n)             (n << 4)
#define MOS_OBJ_HOR(n)            (n << 8)
#define MOS_OBJ_VER(n)            (n << 12)
#define SetMosaic(bh, bv, oh, ov) ((bh) + (bv << 4) + (oh << 8) + (ov << 12))

//////////////////////////////////////////////////////////////////////
// STRUCTURES
////////////////////////////////////////////////////////////////////// 

// dfinit un background compos de tiles

typedef struct t_Background
{
  u16* tileData;                  // donnes des tiles
  u16* mapData;                   // donnes de la map

  u16 size;                       // taille du background

  u8 mosaic,                      // mode mosaque
     colorMode,                   // nombre de couleurs
     number,                      // indice du background (0, 1, 2 ou 3)
     charBaseBlock,               // emplacement mmoire des tiles
     screenBaseBlock,             // emplacement mmoire de la map
     wraparound;                  // mode retour de flot

  s16 xScroll,                    // coordonnes en mode texte
      yScroll,                    // coordonnes en mode texte
      PA,                         // paramtre de rotation
      PB,                         // paramtre de rotation
      PC,                         // paramtre de rotation
      PD;                         // paramtre de rotation

  s32 DX,                         // coordonnes en mode rotation
      DY;                         // coordonnes en mode rotation

} Background, *pBackground;

//////////////////////////////////////////////////////////////////////
// CONSTANTES
////////////////////////////////////////////////////////////////////// 

extern s32 COS[];
extern s32 SIN[];

//////////////////////////////////////////////////////////////////////
// PROTOTYPES
////////////////////////////////////////////////////////////////////// 

void enableBackground(pBackground*);
void updateBackground(pBackground*);

//////////////////////////////////////////////////////////////////////
// IMPLEMENTATIONS
////////////////////////////////////////////////////////////////////// 

//////////////////////////////////////////////////////////////////////
// Active un background et le rend affichable.
//
////////////////////////////////////////////////////////////////////// 

void enableBackground(pBackground bg)
{
  u16 temp;

  bg->tileData = (u16*) CharBaseBlock(bg->charBaseBlock);
  bg->mapData  = (u16*) ScreenBaseBlock(bg->screenBaseBlock);
  temp = bg->size | (bg->charBaseBlock << CHAR_SHIFT) | (bg->screenBaseBlock << SCREEN_SHIFT)
                  | bg->colorMode | bg->mosaic | bg->wraparound;

  switch (bg->number)
  {
    case 0:
      REG_BG0CNT = temp;
      REG_DISPCNT |= BG0_ENABLE;
      break;
    case 1:
      REG_BG1CNT = temp;
      REG_DISPCNT |= BG1_ENABLE;
      break;
    case 2:
      REG_BG2CNT = temp;
      REG_DISPCNT |= BG2_ENABLE;
      break;
    case 3:
      REG_BG3CNT = temp;
      REG_DISPCNT |= BG3_ENABLE;
      break;
  }

} // fin enableBackground()

//////////////////////////////////////////////////////////////////////
// Mets  jour les caractristiques de rotation, de zoom et de
// position d'un background.
//
////////////////////////////////////////////////////////////////////// 

void updateBackground(pBackground bg)
{
  switch (bg->number)
  {
    case 0:
      REG_BG0HOFS = bg->xScroll;
      REG_BG0VOFS = bg->yScroll;
      break;

    case 1:
      REG_BG1HOFS = bg->xScroll;
      REG_BG1VOFS = bg->yScroll;
      break;

    case 2:
      // en mode graphique 0, le background 2 est
      // un background texte
      if (!(REG_DISPCNT & MODE_0))
      {
        REG_BG2X = bg->DX;
        REG_BG2Y = bg->DY;
  
        REG_BG2PA = bg->PA;
        REG_BG2PB = bg->PB;
        REG_BG2PC = bg->PC;
        REG_BG2PD = bg->PD;

      } else  {
        REG_BG2HOFS = bg->xScroll;
        REG_BG2VOFS = bg->yScroll;
      }
      break;

    case 3:
      // en mode graphique 0, le background 3 est
      // un background texte
      if (!(REG_DISPCNT & MODE_0))
      {
        REG_BG3X = bg->DX;
        REG_BG3Y = bg->DY;
  
        REG_BG3PA = bg->PA;
        REG_BG3PB = bg->PB;
        REG_BG3PC = bg->PC;
        REG_BG3PD = bg->PD;

      } else {
        REG_BG3HOFS = bg->xScroll;
        REG_BG3VOFS = bg->yScroll;
      }
      break;
  }

} // fin updateBackground()

//////////////////////////////////////////////////////////////////////
// Mets  jour les caractristiques de rotation, de zoom et de
// position d'un background.
//
////////////////////////////////////////////////////////////////////// 

void rotateBackground(pBackground bg, int angle, int center_x, int center_y, s32 zoom)
{
  center_y = (center_y * zoom) >> 8;
  center_x = (center_x * zoom) >> 8;

  bg->DX = (bg->xScroll << 8) - center_y * SIN[angle] - center_x * COS[angle];
  bg->DY = (bg->yScroll << 8) - center_y * COS[angle] + center_x * SIN[angle];

  bg->PA = ( COS[angle] * zoom) >> 8;
  bg->PB = ( SIN[angle] * zoom) >> 8;
  bg->PC = (-SIN[angle] * zoom) >> 8;
  bg->PD = ( COS[angle] * zoom) >> 8;

} // fin rotateBackground()

#endif

// fin de background.h
