//////////////////////////////////////////////////////////////////////
// INCLUSIONS
//////////////////////////////////////////////////////////////////////
#include "string.h"

#include "main.h"
#include "saving.h"
#include "sfx.h"
#include "sound.h"
#include "sprites.h"
#include "windows.h"

// BACKGROUNDS
#include "background.h"

//////////////////////////////////////////////////////////////////////
// VARIABLES GLOBALES
//////////////////////////////////////////////////////////////////////

#define tiles_WIDTH             64
#define tiles_HEIGHT            64 
#define xwing_WIDTH             64
#define xwing_HEIGHT            64 

#define OFFSET_IS_SAVE          0
#define OFFSET_XWING_X          5
#define OFFSET_XWING_Y          6
#define OFFSET_XWING_ANGLE      7
#define OFFSET_BACKGROUND_X     11
#define OFFSET_BACKGROUND_Y     15

// voir main.h
extern const unsigned char map[];
extern const u16 retaliatorPalette[];
extern const u16 retaliatorData[];
extern const u16 tilesPalette[];
extern const u16 tilesData[];
extern const u16 xwingPalette[];
extern const u16 xwingData[];
extern const u32 _binary_sound_wav_start[];

void interruptProcess(void) __attribute__ ((section(".iwram")));

/////////////////////
// SPRITES
/////////////////////

// voir sprites.h

//////////////////////////////////////////////////////////////////////
// PROTOTYPES
//////////////////////////////////////////////////////////////////////

void displayPicture(const u16[], const u16[]);
void getInput();
void initGame();
void introduction();
void loadState();
void saveState();
void startGame();

//////////////////////////////////////////////////////////////////////
// VARIABLES
//////////////////////////////////////////////////////////////////////

u16 alpha;
pSprite xwing;
pBackground bg2;

//////////////////////////////////////////////////////////////////////
// FONCTIONS
//////////////////////////////////////////////////////////////////////

void interruptProcess()
{
  // n'est utilis que lorsque la lecture d'un son par interruption
  // est active
  REG_TM0CNT  = 0;
  REG_TM1CNT  = 0;
  REG_DMA1CNT = 0;

  REG_IF |= REG_IF;
  REG_IME = 0;
  REG_IE  = 0;
}

void playDirectSound()
{
  REG_SOUNDCNT_X = OPERATE_SOUND_ALL;
  REG_SOUNDCNT_L = SOUNDLEVEL_L(7) | SOUNDLEVEL_R(7);
  REG_SOUNDCNT_H = DSOUND_A_FIFO | DSOUND_A_FULL | DSOUND_A_L_OUTPUT | DSOUND_A_R_OUTPUT;
  
  REG_DMA1SAD = (u32) _binary_sound_wav_start;
  REG_DMA1DAD = 0x40000A0;
  REG_DMA1CNT = DMA_ENABLE | DMA_TIMEING_FIFO | DMA_32 | DMA_REPEAT |
                DMA_SOURCE_INCREMENT | DMA_DEST_FIXED;

  ////////////////////////////////////////////////////////////////////
  // Ce code peut tre dcomment pour arrter la lecture d'un son
  //  l'aide des interruptions
  // REG_TM1D = 0x63C0;
  // REG_TM1CNT = TIMER_CASCADE | TIMER_ENABLE | TIMER_IRQ;
  // 
  // REG_IE  = IRQ_TM1;
  // REG_IME = IRQ_ENABLE;

  REG_TM0D = PLAY_16KHZ;
  REG_TM0CNT = TIMER_ENABLE;
}

int main()
{
  playDirectSound();

  // page d'accueil
  introduction();

  // boucle du jeu
  startGame();

  return 0;

} // fin main()

//////////////////////////////////////////////////////////////////////
// Introduction du jeu
//
//////////////////////////////////////////////////////////////////////

void introduction()
{
  // mode 4
  REG_DISPCNT = MODE_4 | BG2_ENABLE;
  REG_BG2CNT = BG_MOSAIC_ENABLE;

  // fond
  displayPicture(retaliatorData, retaliatorPalette);

  // on attend le fameux "press start"
  while (*KEYS & KEY_START);

  // un petit bruit parce que quand mme hein :-)
  // canal 1
  //  REG_SOUNDCNT_X = OPERATE_SOUND_ALL;
  //  REG_SOUNDCNT_L = SOUND1_L_OUTPUT | SOUND1_R_OUTPUT | SOUNDLEVEL_L(7) | SOUNDLEVEL_R(7);
  //  REG_SOUNDCNT_H = OUTPUT_RATIO_FULL;
  //
  //  REG_SOUND1CNT_L = SOUND1SWEEPSHIFTS(6) | SOUND1SWEEPTIME(5) | SOUND1SWEEPINC;
  //  REG_SOUND1CNT_H = SOUNDDUTY50 | SOUND1ENVDEC | SOUND1ENVINIT(12);
  //  REG_SOUND1CNT_X = 0x0400 | SOUND1INIT;

  // fondu au noir
  // fadeOut();
  s16 mosaic = 0;
  for ( ; mosaic < 16; mosaic++)
  {
    REG_MOSAIC = SetMosaic(mosaic, mosaic, 0, 0);
    waitTime(0, 1000 / 16);
  }

  // un autre petit bruit marrant avec le canal 2
  // REG_SOUNDCNT_X = OPERATE_SOUND_ALL;
  //  REG_SOUNDCNT_L = SOUND2_L_OUTPUT | SOUND2_R_OUTPUT | SOUNDLEVEL_L(7) | SOUNDLEVEL_R(7);
  //  REG_SOUNDCNT_H = OUTPUT_RATIO_FULL;
  //
  //  REG_SOUND2CNT_L = SOUNDDUTY50 | SOUND2ENVDEC | SOUND2ENVINIT(12);
  //  REG_SOUND2CNT_H = 0x0400 | SOUND2INIT | SOUND2PLAYONCE;

} // fin introduction()

//////////////////////////////////////////////////////////////////////
// Lance le jeu
//
//////////////////////////////////////////////////////////////////////

void startGame()
{
  // charge les sprites, etc...
  initGame();

  // alpha blending
  alpha = 8;
  REG_COLEV = BLEND(16, alpha);
  // en spcifiant OBJ comme premire cible de l'alpha blending
  // tous les sprites se verraient affects
  // lorsque TARGET1_OBJ n'est pas prcis, seuls les sprites en
  // MODE_ALPHA sont affects
  REG_BLDMOD = /* TARGET1_OBJ | */ TARGET2_BG2 | SFX_ALPHA;

  for (s16 mosaic = 15; mosaic >= 0; mosaic--)
  {
    REG_MOSAIC = SetMosaic(mosaic, mosaic, 0, 0);
    rotateBackground(bg2, 0, SCREEN_WIDTH >> 1, SCREEN_HEIGHT >> 1, 1 << 8);
    // v-sync
    waitForVSync();
    // mise  jour des backgrounds
    updateBackground(bg2);
  }

  REG_DISPCNT |= OBJ_ENABLE;

  // v-sync
  waitForVSync();
  // copie les sprites
  copyOAMBuffer();

  s32 zoom = 1 << 8;
  s8  inout = -2;

  // boucle principale
  while (1)
  {
    ////////////////////////////////

    // gestion des touches
    getInput();
    // dplace le sprite
    moveSprite(xwing, xwing->x, xwing->y);
    // rotation et mise  l'chelle
    rotateSprite(xwing, xwing->angle, xwing->xscale, xwing->yscale);
    // rotation du fond de l'cran
    rotateBackground(bg2, (xwing->angle == 0 ? 0 : 359 - xwing->angle),
                          SCREEN_WIDTH >> 1, SCREEN_HEIGHT >> 1, /* zoom */ zoom);

    /* un *poil* emmerdant quand on dbug :-))
    zoom += inout;
    if (zoom <= 2)
      inout = 2;
    else if (zoom >= (1 << 8))
      inout = -2;
    */

    // v-sync
    waitForVSync();
    // copie les sprites
    copyOAMBuffer();
    // mise  jour des backgrounds
    updateBackground(bg2);
  }

  delete bg2;
  delete xwing;

} // fin startGame()

//////////////////////////////////////////////////////////////////////
// Gre l'appui sur les touches
//
//////////////////////////////////////////////////////////////////////

void getInput()
{
  // dplacement vers le bas
  if (!(*KEYS & KEY_DOWN))
  {
    if (xwing->y < SCREEN_HEIGHT - xwing_HEIGHT)
      xwing->y++;
    else if (bg2->yScroll + (SCREEN_HEIGHT >> 1) < 512)
      bg2->yScroll++;
  }

  // dplacement vers la droite
  if (!(*KEYS & KEY_RIGHT))
  {
    if (xwing->x < SCREEN_WIDTH - xwing_WIDTH)
      xwing->x++;
    else if (bg2->xScroll + (SCREEN_WIDTH >> 1) < 512)
      bg2->xScroll++;
  }

  // dplacement vers le haut
  if (!(*KEYS & KEY_UP))
  {
    if (xwing->y > 0)
      xwing->y--;
    else if (bg2->yScroll > SCREEN_HEIGHT >> 1)
      bg2->yScroll--;
  }

  // dplacement vers la gauche
  if (!(*KEYS & KEY_LEFT))
  {
    if (xwing->x > 0)
      xwing->x--;
    else if (bg2->xScroll > SCREEN_WIDTH >> 1)
      bg2->xScroll--;
  }

  // on augmente l'opacit
  if (!(*KEYS & KEY_R) && (alpha < 16))
  {
    alpha++;
    REG_COLEV = BLEND(16, alpha);
  }

  // on diminue l'opacit
  if (!(*KEYS & KEY_L) && (alpha > 0))
  {
    alpha--;
    REG_COLEV = BLEND(16, alpha);
  }

  // rotation vers la droite
  if (!(*KEYS & KEY_B))
  {
    xwing->angle++;
    if (xwing->angle >= 360)
      xwing->angle = 0;
  }

  // rotation vers la gauche
  if (!(*KEYS & KEY_A))
  {
    xwing->angle--;
    if (xwing->angle < 0)
      xwing->angle = 359;
  }

  if (!(*KEYS & KEY_START))
  {
    saveState();
  }

} // fin getInput()

//////////////////////////////////////////////////////////////////////
// Sauvegarde l'tat courant du jeu.
//
//////////////////////////////////////////////////////////////////////

void saveState()
{
  saveString(OFFSET_IS_SAVE, "wing");

  saveByte(OFFSET_XWING_X, (u8) xwing->x);
  saveByte(OFFSET_XWING_Y, (u8) xwing->y);

  saveInt(OFFSET_XWING_ANGLE, xwing->angle);

  saveInt(OFFSET_BACKGROUND_X, bg2->xScroll);
  saveInt(OFFSET_BACKGROUND_Y, bg2->yScroll);

} // fin saveState()

//////////////////////////////////////////////////////////////////////
// Charge l'tat du jeu sauvegard.
//
//////////////////////////////////////////////////////////////////////

void loadState()
{
  char* magic = new char[5];
  loadString(OFFSET_IS_SAVE, 4, magic);

  if (strcmp(magic, "wing") == 0)
  {
    xwing->x = (int) loadByte(OFFSET_XWING_X);
    xwing->y = (int) loadByte(OFFSET_XWING_Y);

    xwing->angle = loadInt(OFFSET_XWING_ANGLE);

    bg2->xScroll = loadInt(OFFSET_BACKGROUND_X);
    bg2->yScroll = loadInt(OFFSET_BACKGROUND_Y);
  }

  delete[] magic;

} // fin loadState()

//////////////////////////////////////////////////////////////////////
// Affiche une image plein cran en mode 4.
//
// param : u16[] picture,       l'image
// param : u16[] pictpal,       la palette de couleurs
//
//////////////////////////////////////////////////////////////////////

void displayPicture(const u16 picture[], const u16 pictpal[])
{
  // copie de la palette de couleurs
  // XXX : utilisation du DMA
  for (int i = 0; i < 256; i++)
    palette[i] = pictpal[i];

  // affichage de l'image
  REG_DMA3SAD = (u32) picture;
  REG_DMA3DAD = (u32) videoBuffer;
  REG_DMA3CNT = SCREEN_HEIGHT * SCREEN_WIDTH / 2 | DMA_16NOW;

} // fin displayPicture()

//////////////////////////////////////////////////////////////////////
// Initialise les donnes des tiles et de la map.
//
//////////////////////////////////////////////////////////////////////

void initBackground()
{
  // initialisation du background
  bg2 = new Background;
  bg2->number = 2;
  bg2->charBaseBlock = 3;
  bg2->screenBaseBlock = 31;
  bg2->colorMode = BG_COLOR_256;
  bg2->size = ROTBG_SIZE_512x512;
  bg2->mosaic = BG_MOSAIC_ENABLE;
  bg2->xScroll = SCREEN_WIDTH  >> 1;
  bg2->yScroll = SCREEN_HEIGHT >> 1;

  enableBackground(bg2);

  // copie de la palette des tiles
  for (int i = 0; i < 256; i++)
    palette[i] = tilesPalette[i];

  // copie des donnes des tiles
  REG_DMA3SAD = (u32) tilesData;
  REG_DMA3DAD = (u32) bg2->tileData;
  REG_DMA3CNT = tiles_WIDTH * tiles_HEIGHT / 2 | DMA_16NOW;
  //for (int i = 0; i < tiles_WIDTH * tiles_HEIGHT / 2; i++)
  //  bg2->tileData[i] = tilesData[i];

  u16* temp = (u16 *) map;
  for (int i = 0; i < 64 * 64 / 2; i++)
    bg2->mapData[i] = temp[i];

} // fin initBackground()

//////////////////////////////////////////////////////////////////////
// Initialise les donnes du programme.
//
//////////////////////////////////////////////////////////////////////

void initGame()
{
  // mode 2
  REG_DISPCNT = MODE_2 | BG2_ENABLE | /* OBJ_ENABLE | */ OBJ_MAP_1D;

  // sprites
  // initAngles();
  initSprites();

  // background
  initBackground();

  // palette de couleur des sprites
  for (int i = 0; i < 256; i++)
    spritesPalette[i] = xwingPalette[i];

  // spriteID = 0
  // rotationID = 0
  xwing = createSprite(0, 0);
  xwing->x = 0;
  xwing->y = 0;

  // charge la sauvegarde
  loadState();

  // donnes des sprites
  // en mode bitmap (mode 4),
  // les 512 premiers octets ne sont pas disponibles
  REG_DMA3SAD = (u32) xwingData;
  REG_DMA3DAD = (u32) (&OAMData[512 * 16]);
  // 256 * 8 :
  // nous avons un sprite de 64x64 pixels, divis en tiles
  // de 8x8 pixels, donc nous avons 8x8 tiles.
  // un tiles contient 8x8 pixels
  // 8x8x8 = 512 octets pour une ligne de tiles
  // il y a 8 lignes de tiles
  // comme nous copions 2 octets  la fois, nous copions 256*8
  // fois 16 bits
  REG_DMA3CNT = 256 * 8 | DMA_16NOW;

  //for (int i = xwing->spriteID; i < 256 * 8; i++)
  //  OAMData[i + 512 * 16] = xwingData[i];

  // initialisation du premier sprite
  xwing->oam->attribute0 = COLOR_256 | /* SIZE_DOUBLE | */ MODE_ALPHA |
                           ROTATION_FLAG | SQUARE | xwing->y;
  xwing->oam->attribute1 = SIZE_64 | ROTDATA(xwing->rotationID) | xwing->x;
  xwing->oam->attribute2 = xwing->spriteID + 512; // mode bitmap

} // fin initGame()

// Fin de main.cpp