#ifndef GBA_MAIN_H_
#define GBA_MAIN_H_

//////////////////////////////////////////////////////////////////////
// INCLUSIONS NECESSAIRES
//////////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

//////////////////////////////////////////////////////////////////////
// TYPES ET DEFINITIONS
//////////////////////////////////////////////////////////////////////

// couleur RGB
#define RGB(r, g, b)     ((r) | (g << 5) | (b << 10))

// dimensions de l'cran
#define SCREEN_WIDTH     240
#define SCREEN_HEIGHT    160

// modes
#define MODE_0           0x0
#define MODE_1           0x1
#define MODE_2           0x2
#define MODE_3           0x3
#define MODE_4           0x4
#define MODE_5           0x5

// permet d'activer l'affichage des
// backgrounds et des sprites
#define OBJ_MAP_2D       0x0
#define OBJ_MAP_1D       0x40

#define BG0_ENABLE       0x100
#define BG1_ENABLE       0x200
#define BG2_ENABLE       0x400
#define BG3_ENABLE       0x800
#define OBJ_ENABLE       0x1000

// activation des fentres
#define WIN0_ENABLE      0x2000 
#define WIN1_ENABLE      0x4000
#define WINOBJ_ENABLE    0x8000

// double buffering
#define BACKBUFFER       0x10
#define videoFrontBuffer (u16*) 0x6000000
#define videoBackBuffer  (u16*) 0x600A000

// background 2
#define BG2              0x400;

// touches
#define KEY_A            1
#define KEY_B            2
#define KEY_SELECT       4
#define KEY_START        8
#define KEY_RIGHT        16
#define KEY_LEFT         32
#define KEY_UP           64
#define KEY_DOWN         128
#define KEY_R            256
#define KEY_L            512

// timers
#define REG_TM0D         *(u16*) 0x4000100
#define REG_TM0CNT       *(u16*) 0x4000102
#define REG_TM1D         *(u16*) 0x4000104
#define REG_TM1CNT       *(u16*) 0x4000106
#define REG_TM2D         *(u16*) 0x4000108
#define REG_TM2CNT       *(u16*) 0x400010A
#define REG_TM3D         *(u16*) 0x400010C
#define REG_TM3CNT       *(u16*) 0x400010E

#define FREQUENCY_0      0
#define FREQUENCY_64     1
#define FREQUENCY_256    2
#define FREQUENCY_1024   3

#define TIMER_CASCADE    4
#define TIMER_IRQ        64
#define TIMER_ENABLE     128

// IRQ
#define REG_IE        *(u16*) 0x4000200
#define REG_IF        *(u16*) 0x4000202
#define REG_IME       *(u16*) 0x4000208

#define IRQ_VBLANK    1
#define IRQ_HBLANK    2
#define IRQ_Y_TRIGGER 4
#define IRQ_TM0       8
#define IRQ_TM1       16
#define IRQ_TM2       32
#define IRQ_TM3       64
#define IRQ_COMMS     128
#define IRQ_DMA0      256
#define IRQ_DMA1      512
#define IRQ_DMA2      1024
#define IRQ_DMA3      2048
#define IRQ_KEY       4096
#define IRQ_CART      8192

#define IRQ_ENABLE    1

// cran
#define REG_DISPCNT      *(u32*) 0x4000000

// DMA
#define DMA_ENABLE                    0x80000000
#define DMA_INTERUPT_ENABLE           0x40000000
#define DMA_TIMEING_IMMEDIATE         0x00000000
#define DMA_TIMEING_VBLANK            0x10000000
#define DMA_TIMEING_HBLANK            0x20000000
#define DMA_TIMEING_FIFO              0x30000000
#define DMA_16                        0x00000000
#define DMA_32                        0x04000000
#define DMA_REPEAT                    0x02000000
#define DMA_SOURCE_INCREMENT          0x00000000
#define DMA_SOURCE_DECREMENT          0x00800000
#define DMA_SOURCE_FIXED              0x01000000
#define DMA_DEST_INCREMENT            0x00000000
#define DMA_DEST_DECREMENT            0x00200000
#define DMA_DEST_FIXED                0x00400000
#define DMA_DEST_RELOAD               0x00600000

#define DMA_32NOW                     DMA_ENABLE | DMA_TIMEING_IMMEDIATE | DMA_32 
#define DMA_16NOW                     DMA_ENABLE | DMA_TIMEING_IMMEDIATE | DMA_16 

#define REG_DMA0SAD                   *(u32*) 0x40000B0
#define REG_DMA0DAD                   *(u32*) 0x40000B4
#define REG_DMA0CNT                   *(u32*) 0x40000B8

#define REG_DMA1SAD                   *(u32*) 0x40000BC
#define REG_DMA1DAD                   *(u32*) 0x40000C0
#define REG_DMA1CNT                   *(u32*) 0x40000C4

#define REG_DMA2SAD                   *(u32*) 0x40000C8
#define REG_DMA2DAD                   *(u32*) 0x40000CC
#define REG_DMA2CNT                   *(u32*) 0x40000D0

#define REG_DMA3CNT                   *(u32*) 0x40000DC
#define REG_DMA3DAD                   *(u32*) 0x40000D8
#define REG_DMA3SAD                   *(u32*) 0x40000D4

#define REG_SGFIF0A                   *(u32*) 0x40000A0

// types usuels sur GBA
typedef unsigned char    u8;
typedef unsigned short   u16;
typedef unsigned long    u32;

typedef signed char      s8;
typedef signed short     s16;
typedef signed long      s32;

//////////////////////////////////////////////////////////////////////
// VARIABLES
//////////////////////////////////////////////////////////////////////

// cran
u16* videoBuffer   = (u16*) 0x6000000;;
// palette de couleurs
u16* palette       = (u16*) 0x5000000;
// touches
volatile u32* KEYS = (volatile u32*) 0x04000130;

//////////////////////////////////////////////////////////////////////
// PROTOTYPES
//////////////////////////////////////////////////////////////////////

// fonctions graphiques
void waitForVSync();
void waitTime(int, int);

//////////////////////////////////////////////////////////////////////
// IMPLEMENTATIONS
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Renvoie vrai lorsque la GBA a fini le rafraichissement de l'cran.
// Nous pouvons alors dessiner  nouveau sur l'cran.
//
//////////////////////////////////////////////////////////////////////

void waitForVSync()
{
  volatile u16* vreg = (volatile u16*) 0x04000004;
  while (  *vreg & (1 << 0));
  while (!(*vreg & (1 << 0)));

} // fin waitForVSync()

//////////////////////////////////////////////////////////////////////
// Permet d'interrompre l'excution du programme pendant une dure
// dtermine en secondes et millisecondes.
//
//////////////////////////////////////////////////////////////////////

void waitTime(int seconds, int milliSeconds)
{
  // activation des timers
  REG_TM2CNT = FREQUENCY_256 | TIMER_ENABLE;
  REG_TM3CNT = TIMER_CASCADE | TIMER_ENABLE;

  // on commence  zro
  REG_TM2D = 0;
  REG_TM3D = 0;

  // on attend x secondes
  while (REG_TM3D < seconds);

  // on repart  zro pour les millisecondes
  REG_TM2D = 0;

  // on attend x millisecondes
  while (REG_TM2D / (65536 / 1000) < milliSeconds);

  // on coupe les timers
  REG_TM2CNT = 0;
  REG_TM3CNT = 0;
  
  // on remet  zro
  REG_TM2D = 0;
  REG_TM3D = 0;

} // fin waitTime()

#endif

// fin de main.h