#ifndef _GBA_SPRITES_H
#define _GBA_SPRITES_H

#include <math.h>

//////////////////////////////////////////////////////////////////////
// TYPES ET DEFINITIONS
//////////////////////////////////////////////////////////////////////

#define MAX_SPRITES           128

// attribut numro 0
#define ROTATION_FLAG         0x100
#define SIZE_DOUBLE           0x200
#define MODE_NORMAL           0x0
#define MODE_ALPHA            0x400
#define MODE_WINDOWED         0x800
#define MOSAIC                0x1000
#define COLOR_16              0x0000
#define COLOR_256             0x2000
#define SQUARE                0x0
#define WIDE                  0x4000
#define TALL                  0x8000

// attribut numro 2
#define ROTDATA(n)            (n << 9)
#define HORIZONTAL_FLIP       0x1000
#define VERTICAL_FLIP         0x2000
#define SIZE_8                0x0
#define SIZE_16               0x4000
#define SIZE_32               0x8000
#define SIZE_64               0xC000

// attribut numro 2
#define PRIORITY(n)           ((n) << 10)
#define PALETTE(n)            ((n) << 12)

// rotations
#define PI                    3.14159
#define RADIAN(n)             (((float) n)/ (float) 180 * PI)

//////////////////////////////////////////////////////////////////////
// STRUCTURES
//////////////////////////////////////////////////////////////////////

// dfinit une entit OAM

typedef struct t_OAMEntry
{
  u16 attribute0;
  u16 attribute1;
  u16 attribute2;
  u16 attribute3;

} OAMEntry, *pOAMEntry;

// dfinit un sprite

typedef struct t_Sprite
{
  u16 spriteID;           // identificateur du sprite
  u16 rotationID;         // indicateur de rotation/mise  l'chelle (entre 0 et 31)
  pOAMEntry oam;          // donnes OAM

  int x,                  // coordonne x
      y,                  // coordonne y
      angle;              // angle de rotation
  s32 xscale,             // mise  l'chelle x
      yscale;             // mise  l'chelle x

} Sprite, *pSprite;

// structure de rotations

typedef struct t_RotData
{
  // OAMBuffer[0]
  u16 filler1[3];
  u16 corner1;

  // OAMBuffer[1]
  u16 filler2[3];
  u16 corner2;

  // OAMBuffer[2]
  u16 filler3[3];
  u16 corner3;

  // OAMBuffer[3]
  u16 filler4[3];
  u16 corner4;

} RotData, *pRotData;

//////////////////////////////////////////////////////////////////////
// VARIABLES
//////////////////////////////////////////////////////////////////////

// tableau de sprites
OAMEntry OAMBuffer[MAX_SPRITES];
// mmoire OAM
u16* OAM = (u16*) 0x7000000;
// donnes des sprites
u16* OAMData = (u16*) 0x6010000;
// palette de couleurs des sprites
u16* spritesPalette = (u16*) 0x5000200;

// tables trigonomtriques
//s32 COS[360];
//s32 SIN[360];
extern s32 COS[];
extern s32 SIN[];
// table de rotation
pRotData rotationData = (pRotData) OAMBuffer;

//////////////////////////////////////////////////////////////////////
// PROTOTYPES
//////////////////////////////////////////////////////////////////////

pSprite createSprite(u16, u16);
void    copyOAMBuffer();
void    initAngles();
void    initSprites();
void    moveSprite(pSprite, int, int);
void    rotateSprite(pSprite, s32, int, int);

//////////////////////////////////////////////////////////////////////
// IMPLEMENTATIONS
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Cre un sprite en dehors de l'cran avec un angle de rotation de 0
// et une mise  l'chelle de 1:1.
// Le spriteID correspond  l'entre OAM ddie au sprite.
// Le rotationID correspond  l'entre dans la table de rotation.
//
//////////////////////////////////////////////////////////////////////

pSprite createSprite(u16 spriteID, u16 rotationID)
{
  pSprite sprite = new Sprite;
    // cration du sprite
  sprite->spriteID = spriteID;
  sprite->rotationID = rotationID;

  // informations sur la position,
  // l'orientation et la taille
  sprite->x = 240;
  sprite->y = 160;
  sprite->angle = 0;
  sprite->xscale = (1 << 8);
  sprite->yscale = (1 << 8);

  // entre OAM
  sprite->oam = &OAMBuffer[spriteID];

  return sprite;

} // fin createSprite()

//////////////////////////////////////////////////////////////////////
// Initialise la table trigonomtrique des cosinus et sinus.
//
//////////////////////////////////////////////////////////////////////
/*
void initAngles()
{
  for (int i = 0; i < 360; i++)
  {
    COS[i] = (s32) (cos(RADIAN(i)) * 256);
    SIN[i] = (s32) (sin(RADIAN(i)) * 256);
  }

} // fin initAngles()
*/

//////////////////////////////////////////////////////////////////////
// Provoque la rotation d'un sprite.
//
// param : pSprite lpSprite,      le sprite
// param : int angle,             l'angle de rotation
// param : s32 xscale,            mise  l'chelle en x
// param : s32 yscale,            mise  l'chelle en y
//
//////////////////////////////////////////////////////////////////////

void rotateSprite(pSprite lpSprite, int angle, s32 xscale, s32 yscale)
{
  rotationData[lpSprite->rotationID].corner1 = (u16) (((xscale) *  COS[angle]) >> 8);
  rotationData[lpSprite->rotationID].corner2 = (u16) (((yscale) *  SIN[angle]) >> 8);
  rotationData[lpSprite->rotationID].corner3 = (u16) (((xscale) * -SIN[angle]) >> 8);
  rotationData[lpSprite->rotationID].corner4 = (u16) (((yscale) *  COS[angle]) >> 8);

} // fin rotateSprite()

//////////////////////////////////////////////////////////////////////
// Dplace un sprite aux coordones indiques.
//
// param : pSprite lPSprite,     un pointeur vers un sprite
// param : int x,                la coordonne x
// param : int y,                la coordonne y
//
//////////////////////////////////////////////////////////////////////

void moveSprite(pSprite lpSprite, int x, int y)
{
  lpSprite->oam->attribute0 &= 0xFF00;
  lpSprite->oam->attribute0 |= y;

  lpSprite->oam->attribute1 &= 0xFE00;
  lpSprite->oam->attribute1 |= x;

} // fin moveSprite()

//////////////////////////////////////////////////////////////////////
// Initialise le buffer OAM des sprites en les plaant tous en
// dehors de l'cran.
//
//////////////////////////////////////////////////////////////////////

void initSprites()
{
  for (int i = 0; i < MAX_SPRITES; i++)
  {
    OAMBuffer[i].attribute0 = 160;            // y = 160
    OAMBuffer[i].attribute1 = 240;            // x = 240
    OAMBuffer[i].attribute2 = 0;
  }

} // fin initSprite()

//////////////////////////////////////////////////////////////////////
// Place le buffer OAM dans la mmoire OAM.
//
//////////////////////////////////////////////////////////////////////

void copyOAMBuffer()
{
  REG_DMA3SAD = (u32) OAMBuffer;
  REG_DMA3DAD = (u32) OAM;
  REG_DMA3CNT = MAX_SPRITES * 4 | DMA_16NOW;

} // fin copyOAMBuffer()

#endif

// fin de sprites.h 
